use shtola::json::json;
use shtola::{RefIR, ShFile, Shtola};
use std::fs;
use std::path::{Path, PathBuf};

#[test]
fn read_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/simple");
    s.destination("../fixtures/dest/read");
    let r = s.build().unwrap();
    assert_eq!(r.files.len(), 1);
    let keys: Vec<&PathBuf> = r.files.keys().collect();
    assert_eq!(keys[0].to_str().unwrap(), "hello.txt");
}

#[test]
fn clean_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/simple");
    s.destination("../fixtures/dest/clean");
    s.clean(true);
    fs::create_dir_all("../fixtures/dest/clean").unwrap();
    fs::write("../fixtures/dest/clean/blah.foo", "").unwrap();
    s.build().unwrap();
    let fpath = PathBuf::from("../fixtures/dest/clean/blah.foo");
    assert_eq!(fpath.exists(), false);
}

#[test]
fn write_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/simple");
    s.destination("../fixtures/dest/write");
    s.clean(true);
    let mw = Box::new(|mut ir: RefIR| {
        let file = ir.files.get(&PathBuf::from("hello.txt")).unwrap().clone();
        ir.files.insert(
            "hello.txt".into(),
            ShFile {
                content: "hello".into(),
                frontmatter: file.frontmatter.clone(),
                raw_content: None,
            },
        );
    });
    s.register(mw);
    s.build().unwrap();
    let dpath = PathBuf::from("../fixtures/dest/write/hello.txt");
    assert!(dpath.exists());
    let file = &fs::read(dpath).unwrap();
    let fstring = String::from_utf8_lossy(file);
    assert_eq!(fstring, "hello");
}

#[test]
fn read_works_with_binaries() {
    let mut s = Shtola::new();
    s.source("../fixtures/read_binary");
    s.destination("../fixtures/dest/read_binary");
    let r = s.build().unwrap();
    assert_eq!(r.files.len(), 1);
    let (name, file) = r.files.iter().next().unwrap();
    assert_eq!(name.to_str().unwrap(), "my_binary");
    assert!(file.raw_content.is_some());
}

#[test]
fn write_works_with_binaries() {
    let mut s = Shtola::new();
    s.source("../fixtures/read_binary");
    s.destination("../fixtures/dest/write_binary");
    s.clean(true);
    s.build().unwrap();
    let dpath = PathBuf::from("../fixtures/dest/write_binary/my_binary");
    assert!(dpath.exists());
    let file = fs::read(dpath).unwrap();
    let fres = String::from_utf8(file);
    assert!(fres.is_err());
}

#[test]
fn frontmatter_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/frontmatter");
    s.destination("../fixtures/dest/frontmatter");
    s.clean(true);
    let r = s.build().unwrap();
    let (_, matter_file) = r.files.iter().last().unwrap();
    let frontmatter = matter_file.frontmatter.get("hello").unwrap();
    assert_eq!(frontmatter, "bro");
}

#[test]
fn no_frontmatter_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/frontmatter");
    s.destination("../fixtures/dest/no_frontmatter");
    s.clean(true);
    s.frontmatter(false);
    let r = s.build().unwrap();
    let (_, matter_file) = r.files.iter().last().unwrap();
    assert!(matter_file.frontmatter.is_null());
}

#[test]
fn ignore_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/ignore");
    s.destination("../fixtures/dest/ignore");
    s.ignores(&mut vec!["ignored.md".to_string()]);
    s.clean(true);
    let r = s.build().unwrap();
    assert_eq!(r.files.len(), 1);
    let (path, _) = r.files.iter().last().unwrap();
    assert_eq!(path.to_str().unwrap(), "not_ignored.md");
}

#[test]
fn source_ignore_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/source_ignore");
    s.destination("../fixtures/dest/source_ignore");
    s.source_ignores(Path::new("../fixtures/source_ignore/ignores"))
        .unwrap();
    s.clean(true);
    let r = s.build().unwrap();
    assert_eq!(r.files.len(), 1);
    let (path, _) = r.files.iter().last().unwrap();
    assert_eq!(path.to_str().unwrap(), "two.txt");
}

#[test]
fn metadata_works() {
    let mut s = Shtola::new();
    s.source("../fixtures/simple");
    s.destination("../fixtures/dest/metadata");
    s.clean(true);
    let mw1 = Box::new(|mut ir: RefIR| {
        ir.metadata.insert("test".into(), json!("foo"));
        ir.metadata.insert("test2".into(), json!({"bar": "baz"}));
    });

    let mw2 = Box::new(|mut ir: RefIR| {
        ir.metadata.insert("test".into(), json!(["a", "b", "c"]));
    });

    s.register(mw1);
    s.register(mw2);
    let r = s.build().unwrap();
    assert_eq!(r.metadata.get("test").unwrap(), &json!(["a", "b", "c"]));
    assert_eq!(r.metadata.get("test2").unwrap(), &json!({"bar": "baz"}));
}
